@@ -29,31 +29,30 @@ import tarfile
 import fnmatch
 import zipfile
 import StringIO
+import glob
+from tools import zip_file_buffer
 
-from dric import Downloader
+from dric import Downloader, Repository
 
 SPEC_FILE_NAME = "spec.json"
 SPEC_FILE_VERSION = "0.1"
 
 DEFAULT_CONFIG_FILE = os.path.join(os.path.expanduser("~"), ".me2pack")
 
+ARTIFACT_SPECIFICATION_RE = re.compile(r"(?P<SPEC>(?P<GROUPID>\w(\w|\d|\-|_)*(\.\w(\w|\d|\-|_)*)*){1}:(?P<ARTIFACT>\w+(\w|\d|\_)*){1}:(?P<VERSION>\d+(\.|\d)*){1}(:(?P<CLASSIFIER>\w+))?)")
 VARIABLE_EXPANSION_RE = re.compile(r"(?P<variable>\${(?P<path>((\w|\-)+)(\.(\w|\-)+)*)(\|(?P<transform>\w+))?})")
 VARIABLE_EXPANSION_TRANSFORM_HANDLERS = {'sh': pipes.quote}
 
-class VariableExpansionError(Exception):
-    pass
-
-class PropertyNotFound(VariableExpansionError):
-    pass
-
-class InvalidProperty(VariableExpansionError):
-    pass
-
-class SpecificationError(Exception):
-    pass
-
-class UnsupportedSpecificationError(SpecificationError):
-    pass
+class VariableExpansionError(Exception): pass
+class PropertyNotFound(VariableExpansionError): pass
+class InvalidProperty(VariableExpansionError): pass
+class SpecificationError(Exception): pass
+class UnsupportedSpecificationError(SpecificationError): pass
+class BundleError(Exception): pass
+class BundleFormatError(BundleError): pass
+class BundleIOError(BundleError): pass
+class UnsupportedBundle(BundleError): pass
+class BundleSpecError(BundleError):pass
 
 def expand_variables(input_string, tree, default = None, panic_if_not_found = True, nesting_level = 0, max_nesting_level = 15):
     _nest_level = nesting_level
@@ -115,9 +114,6 @@ def sha1file(fname):
 
     return m.hexdigest()
 
-
-
-
 def validate_spec(spec):
     if "spec-version" not in spec:
         raise SpecificationError("Missing specification")
@@ -126,6 +122,23 @@ def validate_spec(spec):
     if "bundle" not in spec:
         raise SpecificationError("Missing field `bundle`!")
 
+def parse_artifact_spec(spec):
+    """Parse an artifact specification and return the components
+
+    Valid artifact specification:
+     * ro.ieat:test:3.4:x86
+    """
+    match = ARTIFACT_SPECIFICATION_RE.match(spec)
+    if match is None:
+        raise SpecificationError("Unsupport artifact specification")
+    groups = match.groupdict()
+    group_id = groups['GROUPID']
+    artifact_id = groups['ARTIFACT']
+    version = groups['VERSION']
+    classifier = groups['CLASSIFIER']
+    return group_id, artifact_id, version, classifier
+
+
 class PackageBuilder(object):
     log = logging.getLogger("PackageBuilder")
     def __init__(self, config = {}):
@@ -142,6 +155,7 @@ class PackageBuilder(object):
         assemblySpec = self.config["assembly"]
         download_registry = self.config["me2"]["build"]["download"]
         download_dir = expand_variables("${me2.build.download_dir}", self.config)
+        download_dir = os.path.normcase(download_dir)
 
         if not 'fetch' in assemblySpec:
             self.log.debug('Nothing to download')
@@ -159,7 +173,7 @@ class PackageBuilder(object):
                 self.log.error("Artifact %s marked already as downloaded!", name)
                 continue
             url = expand_variables(entry["url"], self.config)
-            destination = expand_variables(entry["destination"], self.config)
+            destination = os.path.normcase(expand_variables(entry["destination"], self.config))
             self.log.info("Downloading %s -> %s", url, destination)
             self.downloader.fetch(url, destination)
             download_registry[name] = {'file': destination}
@@ -173,7 +187,7 @@ class PackageBuilder(object):
         operation = action["operation"]
         input_file = self.config["me2"]["build"]["download"][name]['file']
         if operation == "extract":
-            extract_dir = expand_variables(action['destination'], self.config)
+            extract_dir = os.path.normcase(expand_variables(action['destination'], self.config))
             self.extract(input_file, extract_dir, options = action)
         else:
             self.log.critical("Operation %s is not supported!", operation)
@@ -261,26 +275,40 @@ class PackageBuilder(object):
                 log.critical("Operation %s is not supported", operation)
 
     def saveDescriptor(self):
-        bundle_dir = expand_variables("${me2.build.bundle_dir}", self.config)
-        desc_file = os.path.join(bundle_dir, "buildspec.json")
+        log = self.log.getChild("saveDescriptor")
+        bundle_dir = os.path.normcase(expand_variables("${me2.build.bundle_dir}", self.config))
+        desc_file = os.path.join(bundle_dir, "build_configuration.json")
+        bundle_desc_file = os.path.join(bundle_dir, SPEC_FILE_NAME)
+        log.debug("Saving descriptor to: %s", bundle_desc_file)
+
         self.log.debug("Saving build time descriptor to %s", desc_file)
         with open(desc_file, "w") as f:
             json.dump(self.config, f, sort_keys = True, indent = 4)
+        self.log.debug("Saving configuration to %s", bundle_desc_file)
+        with open(bundle_desc_file, "w") as f:
+            json.dump(self.getBundleConfig(), f, sort_keys = True, indent = 4)
+
+
+    def getBundleConfig(self):
+        config = {"spec-version": SPEC_FILE_VERSION}
+        config["bundle"] = self.config["bundle"]
+        config["configuration"] = self.config["configuration"]
+        return config
 
 
     def createBundle(self):
         log = self.log.getChild("createBundle")
         bundle_name = expand_variables("${me2.bundle_name}", self.config)
-        bundle_file = expand_variables("${me2.build.bundle}", self.config)
-        bundle_dir = expand_variables("${me2.build.bundle_dir}", self.config)
+        bundle_file = os.path.normcase(expand_variables("${me2.build.bundle}", self.config))
+        bundle_dir = os.path.normcase(expand_variables("${me2.build.bundle_dir}", self.config))
         log.info("Destination bundle: %s", bundle_file)
         if os.path.exists(bundle_file):
             log.warn("bundle file exists and it will be removed: %s", bundle_file)
             os.remove(bundle_file)
 
-
         with zipfile.ZipFile(bundle_file, 'w', allowZip64 = True, compression = zipfile.ZIP_DEFLATED) as bundle:
             bundle.comment = "mOSAIC Bundle for %s" % bundle_name
+            log.debug("Creating bundle with files from: %s", bundle_dir)
             manifest = StringIO.StringIO()
             for dirpath, dirnames, filenames in os.walk(bundle_dir):
                 arc_dir = os.path.relpath(os.path.abspath(dirpath), bundle_dir)
@@ -297,7 +325,7 @@ class PackageBuilder(object):
             bundle.writestr("./MANIFEST.mf", manifest.getvalue())
             manifest.close()
 
-        return bundle_file
+        return Bundle(bundle_file)
 
     def build(self):
         self.make_dirs()
@@ -310,13 +338,100 @@ class PackageBuilder(object):
         self.log.info("Saving build time descriptor")
         self.saveDescriptor()
         self.log.info("Creating bundle")
-        self.createBundle()
+        bundles = self.createBundle()
+        return [bundles, ]
 
     def make_dirs(self):
         bundle_dir = expand_variables("${me2.build.bundle_dir}", self.config)
+        bundle_dir = os.path.normcase(bundle_dir)
         if not os.path.exists(bundle_dir):
             os.makedirs(bundle_dir)
 
+class BundleInfo(object):
+    specification = None
+    group_id = None
+    package_id = None
+    version = None
+    type = None
+    classifier = None
+
+class Bundle(object):
+    log = logging.getLogger("Bundle")
+    info = None
+    def __init__(self, *args, **argv):
+        log = self.log.getChild("__init__")
+        if self.info is not None:
+            return
+        if 'package_file' in argv: package_file = argv['package_file']
+        elif len(args) >= 1: package_file = args[0]
+        else: raise AttributeError("Missing package_file attribute")
+
+        if isinstance(package_file, basestring):
+            self.package_file_path = package_file
+        else:
+            self.package_file_path = package_file.name
+        log.debug("Opening bundle file %s", os.path.abspath(self.package_file_path))
+        self.package_file = zipfile.ZipFile(package_file)
+        self.info = self._get_info()
+        specialized_class = None
+        for klass in Bundle.__subclasses__():
+            if klass.type == self.info.type:
+                specialized_class = klass
+        if specialized_class is None:
+            raise UnsupportedBundle("The bundle type %s" % self.info.type)
+        self.__class__ = specialized_class
+
+        specialized_class.__init__(self, package_file, *args, **argv)
+
+    def _get_info(self):
+        bi = BundleInfo()
+        spec = json.load(self.package_file.open(SPEC_FILE_NAME))
+        bundle = spec['bundle']
+        bi.configuration = spec['configuration']
+        bi.identification = ":".join((bundle['group-id'], bundle['package-id'], bundle['version']))
+        if 'classifier' in bundle:
+            bi.identification += ":%s" % bundle['classifier']
+        bi.type = bundle["type"]
+        bi.group_id = bundle['group-id']
+        bi.package_id = bundle['package-id']
+        bi.version = bundle['version']
+        bi.classifier = bundle.get("classifier", None)
+        bi.bundle_info = bundle
+        return bi
+
+    def extract(self, destination):
+        raise NotImplementedError()
+
+class ContainerBundle(Bundle):
+    type = "container-bundle"
+    log = logging.getLogger("ContainerBundle")
+
+    def extract(self, destination):
+        self.log.debug("Extracting into: %s", destination)
+        self.extract_rootfs(os.path.join(destination, "rootfs"))
+        self.extract_meta(destination)
+        return self
+
+    def extract_meta(self, destination):
+        self.log.debug("Saving metadata to %s", destination)
+        configuration_file = os.path.join(destination, "configuration.json")
+        bundle_file = os.path.join(destination, "bundle.json")
+        with open(configuration_file, "w") as f:
+            json.dump(self.info.configuration, f)
+        with open(bundle_file, "w") as f:
+            json.dump(self.info.bundle_info, f)
+
+    def extract_rootfs(self, destination):
+        self.log.debug("Extracting rootfs into: %s", destination)
+        os.makedirs(destination)
+        tar_object = zip_file_buffer(open(self.package_file_path), "rootfs")
+        tar = tarfile.open(fileobj = tar_object, mode = "r|")
+        tar.extractall(destination)
+        self.log.debug("Finished extracting!")
+
+
+
+
 class Packager(object):
     log = logging.getLogger("Packager")
     def __init__(self, default_config = DEFAULT_CONFIG_FILE):
@@ -327,8 +442,13 @@ class Packager(object):
             self.log.debug("Starting with empty config")
             self.config = {'me2': {}}
 
+
         self.config_file = os.path.join(os.getcwd(), SPEC_FILE_NAME)
-        self.config.update(json.load(open(self.config_file)))
+        try:
+            self.config.update(json.load(open(self.config_file)))
+        except ValueError,e:
+            self.log.warn("Failed parsing spec file from %s", self.config_file)
+            raise BundleSpecError("Failed parsing spec file")
         me2_config = self.config["me2"]
         me2_config['cwd'] = os.getcwd()
         me2_config['bundle_name'] = "${bundle.package-id}-${bundle.version}-${bundle.classifier}"
@@ -344,6 +464,8 @@ class Packager(object):
 
         build_parser = subparsers.add_parser('build', help = 'Validate a package')
         build_parser.add_argument("--buildDir", type = str, default = "${me2.cwd}/me2-build")
+        build_parser.add_argument("--upload", action = 'store_true', default = False)
+        build_parser.add_argument("--repository", type = str, default = "http://developers.mosaic-cloud.eu/artifactory/mosaic")
         build_parser.set_defaults(package_command = self.build)
 
         extract_parser = subparsers.add_parser('extract', help = 'extract a package')
@@ -356,6 +478,13 @@ class Packager(object):
     def extract(self, options):
         raise NotImplementedError()
 
+    def upload(self, repository, bundles):
+        self.log.debug("Starting upload")
+        repo = Repository("upload", repository)
+        for bundle in bundles:
+            repo.register(bundle)
+
+
     def build(self, options):
         config = copy.deepcopy(self.config)
         if 'build' in config['me2']:
@@ -374,7 +503,9 @@ class Packager(object):
             build_config["download"] = {}
 
         builder = PackageBuilder(config = config)
-        builder.build()
+        bundles = builder.build()
+        if options.upload:
+            self.upload(options.repository, bundles)
 
     def handle(self, options):
         options.package_command(options)
